//! This module provides autogeneration of clap commands to manipulate entities that are stored in
//! a microrm database.
//!
//! The main motivations of this module are:
//! - Providing an introspection interface for debugging.
//! - Providing an admin tool for services built on microrm.
//!
//! Broadly speaking, this module aims to play nicely with whatever other clap interface has been
//! set up. The main entry point of interest is [`ClapInterface`], which is a clap `Subcommand`
//! that can be included in an existing derived clap-parsed structure or enum; it can also be
//! included into a clap parse tree if you are using clap's builder interfaces.
//!
//! Once a [`ClapInterface`] object has been constructed, the [`ClapInterface::perform`] method
//! will execute the constructed action, possibly deferring to a user-defined set of custom
//! per-`CLIObject` commands.

use crate::{
    prelude::{Insertable, Queryable},
    schema::entity::Entity,
    Error,
};

mod clap_interface;

pub use clap_interface::{ClapInterface, InterfaceCustomization, ValueRole};

/// Trait used to expose errors from the autogenerated clap interface code. Implemented for
/// [`Error`] but can be implemented for other error types if extra information needs to be passed
/// back.
pub trait CLIError: From<Error> {
    /// Complain about an entity by a given list of members not existing.
    fn no_such_entity(ename: &'static str, query: String) -> Self;
}

impl CLIError for Error {
    fn no_such_entity(_ename: &'static str, _query: String) -> Self {
        Error::EmptyResult
    }
}

/// Extension trait on top of [`Entity`] that provides necessary functionality to interact with
/// instances from the command-line, such as the data required to create an object.
pub trait CLIObject: Sized + Entity + std::fmt::Debug {
    /// What kind of error type should be generated when issues are encountered.
    type Error: CLIError;
    /// Extra data to pass into methods.
    type CommandData;

    /// Required data to create an instance of `Self`.
    type CreateParameters: clap::Parser + std::fmt::Debug;
    /// Create an instance of `Self` from a `CommandData` instance and a `CreateParameters` instance.
    fn create_from_params(
        _: &Self::CommandData,
        _: &Self::CreateParameters,
    ) -> Result<Self, Self::Error>;

    /// Subcommand type representing all custom operations on this CLIObject implementation.
    type ExtraCommands: clap::Subcommand + std::fmt::Debug;
    /// Function that handles a constructed `Self::ExtraCommands` instance from clap.
    fn run_extra_command(
        _data: &Self::CommandData,
        _extra: &Self::ExtraCommands,
        _query_ctx: impl Queryable<EntityOutput = Self>,
        _insert_ctx: &impl Insertable<Self>,
    ) -> Result<(), Self::Error> {
        unreachable!()
    }

    /// A short string used for display purposes to describe this object. Can be the output of
    /// Debug or empty.
    fn shortname(&self) -> String;
}

/// Empty subcommand, used as a default value for CLIObjects that have no implemented additional
/// commands (i.e. the `ExtraCommands` associated type).
#[derive(Debug)]
pub struct EmptyCommand;

impl clap::FromArgMatches for EmptyCommand {
    fn from_arg_matches(_matches: &clap::ArgMatches) -> Result<Self, clap::Error> {
        Err(clap::Error::new(clap::error::ErrorKind::UnknownArgument))
    }

    fn update_from_arg_matches(&mut self, _matches: &clap::ArgMatches) -> Result<(), clap::Error> {
        Ok(())
    }
}

impl clap::Subcommand for EmptyCommand {
    fn augment_subcommands(cmd: clap::Command) -> clap::Command {
        cmd
    }

    fn augment_subcommands_for_update(cmd: clap::Command) -> clap::Command {
        cmd
    }

    fn has_subcommand(_name: &str) -> bool {
        false
    }
}

#[cfg(test)]
mod tests {
    use super::{CLIObject, ClapInterface, InterfaceCustomization, ValueRole};
    use clap::Parser;
    use microrm::prelude::*;
    use test_log::test;

    struct CTRelation;
    impl Relation for CTRelation {
        type Range = Customer;
        type Domain = Transaction;
        const NAME: &'static str = "CTRelation";
    }

    struct ETRelation;
    impl Relation for ETRelation {
        type Range = Employee;
        type Domain = Transaction;
        const NAME: &'static str = "ETRelation";
    }

    #[derive(Entity)]
    struct Customer {
        #[key]
        name: String,

        txs: microrm::RelationRange<CTRelation>,
    }

    #[derive(Entity)]
    struct Employee {
        #[key]
        name: String,

        txs: microrm::RelationRange<ETRelation>,
    }

    #[derive(Entity)]
    struct Transaction {
        #[key]
        title: String,
        amount: isize,

        customer: microrm::RelationDomain<CTRelation>,
        employee: microrm::RelationDomain<ETRelation>,
    }

    #[derive(Database)]
    struct TransactionTestDB {
        customers: microrm::IDMap<Customer>,
        employees: microrm::IDMap<Employee>,
        transactions: microrm::IDMap<Transaction>,
    }

    #[derive(clap::Parser, Debug)]
    struct PCreate {
        name: String,
    }

    impl CLIObject for Customer {
        type Error = microrm::Error;
        type CreateParameters = PCreate;
        fn create_from_params(
            _extra: &(),
            create: &Self::CreateParameters,
        ) -> Result<Self, Self::Error> {
            Ok(Customer {
                name: create.name.clone(),
                txs: Default::default(),
            })
        }
        type ExtraCommands = super::EmptyCommand;
        type CommandData = ();

        fn shortname(&self) -> String {
            self.name.clone()
        }
    }

    impl CLIObject for Employee {
        type Error = microrm::Error;
        type CreateParameters = PCreate;
        fn create_from_params(
            _extra: &(),
            create: &Self::CreateParameters,
        ) -> Result<Self, Self::Error> {
            Ok(Employee {
                name: create.name.clone(),
                txs: Default::default(),
            })
        }
        type ExtraCommands = super::EmptyCommand;
        type CommandData = ();

        fn shortname(&self) -> String {
            self.name.clone()
        }
    }

    #[derive(clap::Parser, Debug)]
    struct TCreate {
        title: String,
        amount: isize,
    }

    impl CLIObject for Transaction {
        type Error = microrm::Error;
        type CreateParameters = TCreate;
        fn create_from_params(
            _: &Self::CommandData,
            cp: &Self::CreateParameters,
        ) -> Result<Self, Self::Error> {
            Ok(Transaction {
                title: cp.title.clone(),
                amount: cp.amount,
                customer: Default::default(),
                employee: Default::default(),
            })
        }

        type ExtraCommands = super::EmptyCommand;
        type CommandData = ();

        fn shortname(&self) -> String {
            String::new()
        }
    }

    #[derive(Debug)]
    struct TestCustomization {}

    impl TestCustomization {
        fn new() -> Self {
            Self {}
        }
    }

    impl InterfaceCustomization for TestCustomization {
        fn has_value_for(entity: &'static str, field: &'static str, role: ValueRole) -> bool {
            if entity == "node" && field == "parent" && role == ValueRole::BaseTarget {
                true
            } else {
                false
            }
        }

        fn value_for(
            &self,
            _entity: &'static str,
            _field: &'static str,
            _role: super::ValueRole,
        ) -> String {
            String::from("")
        }
    }

    #[derive(Debug, clap::Parser)]
    enum Params {
        Customer {
            #[clap(subcommand)]
            cmd: ClapInterface<Customer, TestCustomization>,
        },
        Employee {
            #[clap(subcommand)]
            cmd: ClapInterface<Employee, TestCustomization>,
        },
        Tx {
            #[clap(subcommand)]
            cmd: ClapInterface<Transaction, ()>,
        },
    }

    fn run_cmd(db: &TransactionTestDB, c: &TestCustomization, args: &[&str]) {
        match <Params as Parser>::try_parse_from(args) {
            Ok(Params::Customer { cmd }) => {
                cmd.perform(&(), c, &db.customers, &db.customers)
                    .expect("couldn't perform command");
            },
            Ok(Params::Employee { cmd }) => {
                cmd.perform(&(), c, &db.employees, &db.employees)
                    .expect("couldn't perform command");
            },
            Ok(Params::Tx { cmd }) => {
                cmd.perform(&(), &(), &db.transactions, &db.transactions)
                    .expect("couldn't perform command");
            },
            Err(e) => {
                println!("{}", e.render());
                panic!("error parsing arguments")
            },
        }
    }

    #[test]
    fn simple_entity_create_delete() {
        let db = TransactionTestDB::open_path(":memory:").unwrap();

        assert_eq!(
            db.customers
                .keyed("a_key")
                .count()
                .expect("couldn't count entries"),
            0
        );
        run_cmd(
            &db,
            &TestCustomization::new(),
            &["execname", "customer", "create", "a_key"],
        );
        assert_eq!(
            db.customers
                .keyed("a_key")
                .count()
                .expect("couldn't count entries"),
            1
        );
        run_cmd(
            &db,
            &TestCustomization::new(),
            &["execname", "customer", "delete", "a_key"],
        );
        assert_eq!(
            db.customers
                .keyed("a_key")
                .count()
                .expect("couldn't count entries"),
            0
        );
    }

    #[test]
    fn create_and_attach() {
        let db = TransactionTestDB::open_path(":memory:").unwrap();

        let tc = TestCustomization::new();

        run_cmd(&db, &tc, &["execname", "customer", "create", "cname"]);
        run_cmd(&db, &tc, &["execname", "employee", "create", "ename"]);
        run_cmd(&db, &tc, &["execname", "tx", "create", "tname", "100"]);
        run_cmd(
            &db,
            &tc,
            &["execname", "customer", "attach", "cname", "txs", "tname"],
        );
        run_cmd(
            &db,
            &tc,
            &["execname", "employee", "attach", "ename", "txs", "tname"],
        );

        assert_eq!(
            db.customers
                .keyed("cname")
                .join(Customer::Txs)
                .join(Transaction::Employee)
                .first()
                .get()
                .expect("couldn't run query")
                .expect("no such employee")
                .name,
            "ename"
        );
    }
}
