use crate::{DBResult, Error};
use libsqlite3_sys as sq;
use std::{
    cell::Cell,
    collections::HashMap,
    ffi::{CStr, CString},
    pin::Pin,
    sync::{Arc, Mutex},
};

fn check_rcode<'a>(sql: impl FnOnce() -> Option<&'a str>, rcode: i32) -> Result<(), Error> {
    if rcode == sq::SQLITE_OK {
        Ok(())
    } else {
        Err(Error::Sqlite {
            code: rcode,
            msg: unsafe { CStr::from_ptr(sq::sqlite3_errstr(rcode)) }
                .to_str()?
                .to_string(),
            sql: sql().map(String::from),
        })
    }
}

struct ConnectionData {
    sqlite: *mut sq::sqlite3,
    stmts: HashMap<u64, Statement>,
}

impl Drop for ConnectionData {
    fn drop(&mut self) {
        self.stmts.clear();
        unsafe {
            sq::sqlite3_close(self.sqlite);
        }
    }
}

unsafe impl Send for ConnectionData {}

pub(crate) trait PreparedKey {
    fn into_u64(self) -> u64;
}

impl PreparedKey for u64 {
    fn into_u64(self) -> u64 {
        self
    }
}

impl PreparedKey for std::any::TypeId {
    fn into_u64(self) -> u64 {
        use std::hash::Hash;
        use std::hash::Hasher;
        let mut hasher = std::collections::hash_map::DefaultHasher::new();
        self.hash(&mut hasher);
        hasher.finish()
    }
}

/// Represents a single sqlite connection, in SQLITE_MULTITHREADED mode.
///
/// This translates to a struct that is Send, but not Sync.
#[derive(Clone)]
pub struct Connection(Arc<Mutex<ConnectionData>>);

impl Connection {
    /// Establish a new connection to a sqlite database object. Note that this type carries no schema information, unlike [`Database`](../schema/traits.Database.html).
    pub fn new(url: &str) -> Result<Self, Error> {
        let db_ptr = unsafe {
            let url = CString::new(url)?;
            let mut db_ptr = std::ptr::null_mut();
            check_rcode(
                || None,
                sq::sqlite3_open_v2(
                    url.as_ptr(),
                    &mut db_ptr,
                    sq::SQLITE_OPEN_READWRITE | sq::SQLITE_OPEN_NOMUTEX | sq::SQLITE_OPEN_CREATE,
                    std::ptr::null_mut(),
                ),
            )?;
            db_ptr
        };

        if db_ptr.is_null() {
            return Err(Error::InternalError(
                "sqlite3_open_v2 returned a NULL connection",
            ));
        }

        unsafe {
            sq::sqlite3_busy_timeout(db_ptr, 1000);
        }

        Ok(Self(Arc::new(Mutex::new(ConnectionData {
            sqlite: db_ptr,
            stmts: Default::default(),
        }))))
    }

    /// Execute a raw SQL statement on the database this connection represents. Use with care.
    pub fn execute_raw_sql(&self, sql: impl AsRef<str>) -> DBResult<()> {
        let data = self.0.lock()?;

        log::trace!("executing raw sql: {sql}", sql = sql.as_ref());

        unsafe {
            let c_sql = CString::new(sql.as_ref())?;
            let mut err = std::ptr::null_mut();
            let rcode = sq::sqlite3_exec(
                data.sqlite,
                c_sql.as_ptr(),
                None,
                std::ptr::null_mut(),
                &mut err,
            );

            // special error handling because of the err string
            if rcode != sq::SQLITE_OK {
                let e = Error::Sqlite {
                    code: rcode,
                    msg: if err.is_null() {
                        CStr::from_ptr(sq::sqlite3_errstr(rcode))
                    } else {
                        CStr::from_ptr(err)
                    }
                    .to_str()?
                    .to_string(),
                    sql: Some(sql.as_ref().into()),
                };
                if !err.is_null() {
                    sq::sqlite3_free(err.cast());
                }
                return Err(e);
            }
        }

        Ok(())
    }

    pub(crate) fn with_prepared<R>(
        &self,
        hash_key: impl PreparedKey,
        build_query: impl Fn() -> String,
        run_query: impl Fn(StatementContext) -> DBResult<R>,
    ) -> DBResult<R> {
        let mut data = self.0.lock()?;
        let conn = data.sqlite;

        use std::collections::hash_map::Entry;
        match data.stmts.entry(hash_key.into_u64()) {
            Entry::Vacant(e) => {
                let sql = build_query();

                log::trace!("preparing query: {sql}");

                // prepare the statement
                let mut stmt = std::ptr::null_mut();
                unsafe {
                    check_rcode(
                        || Some(sql.as_str()),
                        sq::sqlite3_prepare_v2(
                            conn,
                            sql.as_ptr().cast(),
                            sql.len() as i32,
                            &mut stmt,
                            std::ptr::null_mut(),
                        ),
                    )?;
                };

                if stmt.is_null() {
                    return Err(Error::InternalError(
                        "sqlite3_prepare_v2 returned a NULL stmt",
                    ));
                }

                let stmt = e.insert(Statement { sqlite: conn, stmt });

                run_query(stmt.make_context()?)
            },
            Entry::Occupied(mut e) => run_query(e.get_mut().make_context()?),
        }
    }
}

pub(crate) struct Transaction<'l> {
    db: &'l Connection,
    committed: bool,
}

impl<'l> Transaction<'l> {
    pub fn new(db: &'l Connection) -> DBResult<Self> {
        db.execute_raw_sql("BEGIN TRANSACTION")?;
        Ok(Self {
            db,
            committed: false,
        })
    }

    pub fn commit(mut self) -> DBResult<()> {
        self.committed = true;

        self.db.execute_raw_sql("COMMIT")
    }
}

impl<'l> Drop for Transaction<'l> {
    fn drop(&mut self) {
        if !self.committed {
            let _ = self.db.execute_raw_sql("ROLLBACK");
        }
    }
}

struct Statement {
    #[allow(unused)]
    sqlite: *mut sq::sqlite3,
    stmt: *mut sq::sqlite3_stmt,
}

impl Statement {
    fn make_context(&mut self) -> DBResult<StatementContext> {
        // begin by resetting the statement
        unsafe {
            check_rcode(|| None, sq::sqlite3_reset(self.stmt))?;
        }

        Ok(StatementContext {
            stmt: self,
            owned_strings: Default::default(),
            done: false.into(),
        })
    }
}

impl Drop for Statement {
    fn drop(&mut self) {
        unsafe {
            sq::sqlite3_finalize(self.stmt);
        }
    }
}

#[cfg(test)]
mod test {
    use super::Connection;

    #[test]
    fn simple_sql() {
        let c = Connection::new(":memory:").expect("couldn't open test db");
        c.execute_raw_sql("CREATE TABLE test_table (id integer primary key, value string)")
            .expect("couldn't execute sql");
    }

    #[test]
    fn prepare_stmt() {
        let c = Connection::new(":memory:").expect("couldn't open test db");
        c.execute_raw_sql("CREATE TABLE test_table (id integer primary key, value string)")
            .expect("couldn't execute sql");

        c.with_prepared(
            1,
            || format!("INSERT INTO test_table VALUES (?, ?)"),
            |ctx| {
                ctx.bind(1, 1usize)?;
                ctx.bind(2, "value")?;

                ctx.iter().last();

                Ok(())
            },
        )
        .expect("couldn't run prepared INSERT statement");

        c.with_prepared(
            2,
            || format!("SELECT * FROM test_table"),
            |ctx| {
                let count = ctx
                    .iter()
                    .map(|row| {
                        let row = row.unwrap();
                        assert_eq!(row.read::<i64>(0).expect("couldn't read row ID"), 1);
                        assert_eq!(
                            row.read::<String>(1).expect("couldn't read row value"),
                            "value"
                        );
                    })
                    .count();
                assert!(count > 0);

                Ok(())
            },
        )
        .expect("couldn't run prepared SELECT statement");
    }
}

/// A table row that an `Entity` can be read out of.
pub struct StatementRow<'a> {
    stmt: &'a Statement,
    _ctx: Option<StatementContext<'a>>,
}

impl<'a> StatementRow<'a> {
    /// Read a single [`Readable`] out of the database row.
    pub fn read<T: Readable>(&self, index: i32) -> DBResult<T> {
        T::read_from(self, index)
    }
}

/// A query context.
pub struct StatementContext<'a> {
    stmt: &'a Statement,
    owned_strings: Vec<Pin<String>>,
    done: Cell<bool>,
}

impl<'a> StatementContext<'a> {
    /// Bind data to the context at a given index.
    ///
    /// Remember that bind indices start at 1!
    pub fn bind<B: Bindable>(&self, index: i32, bindable: B) -> DBResult<()> {
        bindable.bind(self, index)
    }

    /// Transfer ownership of a string to the statement context. For transient strings, this is required in order to have
    /// the string outlive the query.
    pub fn transfer(&mut self, s: Pin<String>) {
        self.owned_strings.push(s);
    }

    fn step(&self) -> DBResult<bool> {
        if self.done.get() {
            return Ok(false);
        }
        match unsafe { sq::sqlite3_step(self.stmt.stmt) } {
            sq::SQLITE_ROW => Ok(true),
            sq::SQLITE_DONE => {
                self.done.set(true);
                Ok(false)
            },
            sq::SQLITE_BUSY => {
                log::trace!("Concurrent database access!");
                todo!()
            },
            sq::SQLITE_CONSTRAINT => {
                let msg = unsafe { CStr::from_ptr(sq::sqlite3_errmsg(self.stmt.sqlite)) }
                    .to_str()
                    .unwrap()
                    .to_string();
                log::trace!("SQLite constraint violation: {msg}");
                Err(Error::ConstraintViolation(msg))
            },
            err => {
                log::trace!("unexpected error during sqlite3_step: {:?}", err);
                check_rcode(|| None, err)?;
                unreachable!()
            },
        }
    }

    // this needs to be replaced with a "single" version that keeps the StatementContext alive, or
    // StatementRow needs an optional StatementContext to keep alive
    #[doc(hidden)]
    pub fn run(self) -> DBResult<Option<StatementRow<'a>>> {
        if self.step()? {
            Ok(Some(StatementRow {
                stmt: self.stmt,
                _ctx: Some(self),
            }))
        } else {
            Ok(None)
        }
    }

    #[doc(hidden)]
    pub fn iter(self) -> impl Iterator<Item = DBResult<StatementRow<'a>>> {
        struct I<'a>(StatementContext<'a>);

        impl<'a> Iterator for I<'a> {
            type Item = DBResult<StatementRow<'a>>;

            fn next(&mut self) -> Option<Self::Item> {
                // XXX: unwrap
                match self.0.step() {
                    Ok(true) => Some(Ok(StatementRow {
                        _ctx: None,
                        stmt: self.0.stmt,
                    })),
                    Ok(false) => None,
                    Err(e) => Some(Err(e)),
                }
                /*map(|_| )*/
            }
        }

        I(self)
    }
}

impl<'a> Drop for StatementContext<'a> {
    fn drop(&mut self) {
        // attempt to bind NULLs into each parameter
        unsafe {
            // clear out the rest of the rows
            while self.step().is_ok_and(|v| v) {}
            sq::sqlite3_clear_bindings(self.stmt.stmt);
        }
    }
}

/// Data that can be bound into a [`StatementContext`].
pub trait Bindable {
    /// Bind data into the [`StatementContext`].
    fn bind<'ctx, 'data: 'ctx>(
        &'data self,
        ctx: &StatementContext<'ctx>,
        index: i32,
    ) -> DBResult<()>;
}

impl Bindable for () {
    fn bind<'ctx, 'data: 'ctx>(
        &'data self,
        ctx: &StatementContext<'ctx>,
        index: i32,
    ) -> DBResult<()> {
        unsafe { check_rcode(|| None, sq::sqlite3_bind_null(ctx.stmt.stmt, index)) }
    }
}

impl Bindable for i64 {
    fn bind<'ctx, 'data: 'ctx>(
        &'data self,
        ctx: &StatementContext<'ctx>,
        index: i32,
    ) -> DBResult<()> {
        unsafe { check_rcode(|| None, sq::sqlite3_bind_int64(ctx.stmt.stmt, index, *self)) }
    }
}

impl Bindable for usize {
    fn bind<'ctx, 'data: 'ctx>(
        &'data self,
        ctx: &StatementContext<'ctx>,
        index: i32,
    ) -> DBResult<()> {
        (*self as i64).bind(ctx, index)
    }
}

impl Bindable for f32 {
    fn bind<'ctx, 'data: 'ctx>(&self, ctx: &StatementContext<'ctx>, index: i32) -> DBResult<()> {
        (*self as f64).bind(ctx, index)
    }
}

impl Bindable for f64 {
    fn bind<'ctx, 'data: 'ctx>(&self, ctx: &StatementContext<'ctx>, index: i32) -> DBResult<()> {
        unsafe {
            check_rcode(
                || None,
                sq::sqlite3_bind_double(ctx.stmt.stmt, index, *self),
            )
        }
    }
}

impl<'a> Bindable for &'a str {
    fn bind<'ctx, 'data: 'ctx>(
        &'data self,
        ctx: &StatementContext<'ctx>,
        index: i32,
    ) -> DBResult<()> {
        unsafe {
            check_rcode(
                || None,
                sq::sqlite3_bind_text(
                    ctx.stmt.stmt,
                    index,
                    self.as_ptr().cast(),
                    self.len() as i32,
                    sq::SQLITE_STATIC(),
                ),
            )
        }
    }
}

impl Bindable for str {
    fn bind<'ctx, 'data: 'ctx>(
        &'data self,
        ctx: &StatementContext<'ctx>,
        index: i32,
    ) -> DBResult<()> {
        <&'_ str>::bind(&self, ctx, index)
    }
}

impl Bindable for String {
    fn bind<'ctx, 'data: 'ctx>(
        &'data self,
        ctx: &StatementContext<'ctx>,
        index: i32,
    ) -> DBResult<()> {
        self.as_str().bind(ctx, index)
    }
}

impl<'a> Bindable for &'a [u8] {
    fn bind<'ctx, 'data: 'ctx>(
        &'data self,
        ctx: &StatementContext<'ctx>,
        index: i32,
    ) -> DBResult<()> {
        unsafe {
            check_rcode(
                || None,
                sq::sqlite3_bind_blob64(
                    ctx.stmt.stmt,
                    index,
                    self.as_ptr().cast(),
                    self.len() as u64,
                    sq::SQLITE_STATIC(),
                ),
            )
        }
    }
}

/// Data that can be read from a [`StatementRow`].
pub trait Readable: Sized {
    /// Read an instance of `Self` from a [`StatementRow`].
    ///
    /// Remember that read indices start at 0!
    fn read_from(sr: &StatementRow<'_>, index: i32) -> DBResult<Self>;
}

/// Readable that can be used to pre-determine if a value is NULL, i.e. for `Option::None`
/// checking.
pub struct IsNull(pub bool);

// NULL-checker
impl Readable for IsNull {
    fn read_from(sr: &StatementRow<'_>, index: i32) -> DBResult<Self> {
        let column_type = unsafe { sq::sqlite3_column_type(sr.stmt.stmt, index) };
        if column_type == sq::SQLITE_NULL {
            Ok(IsNull(true))
        } else {
            Ok(IsNull(false))
        }
    }
}

impl Readable for i64 {
    fn read_from(sr: &StatementRow<'_>, index: i32) -> DBResult<Self> {
        unsafe { Ok(sq::sqlite3_column_int64(sr.stmt.stmt, index)) }
    }
}

impl Readable for usize {
    fn read_from(sr: &StatementRow<'_>, index: i32) -> DBResult<Self> {
        i64::read_from(sr, index).map(|i| i as usize)
    }
}

impl Readable for f32 {
    fn read_from(sr: &StatementRow<'_>, index: i32) -> DBResult<Self> {
        f64::read_from(sr, index).map(|i| i as f32)
    }
}

impl Readable for f64 {
    fn read_from(sr: &StatementRow<'_>, index: i32) -> DBResult<Self> {
        unsafe { Ok(sq::sqlite3_column_double(sr.stmt.stmt, index)) }
    }
}

impl Readable for String {
    fn read_from(sr: &StatementRow<'_>, index: i32) -> DBResult<Self> {
        unsafe {
            let text = sq::sqlite3_column_text(sr.stmt.stmt, index);
            if text.is_null() {
                Err(Error::InternalError(
                    "NULL pointer result from sqlite3_column_text",
                ))
            } else {
                Ok(CStr::from_ptr(text.cast()).to_str()?.to_string())
            }
        }
    }
}

impl Readable for Vec<u8> {
    fn read_from(sr: &StatementRow<'_>, index: i32) -> DBResult<Self> {
        unsafe {
            let ptr = sq::sqlite3_column_blob(sr.stmt.stmt, index);
            let len = sq::sqlite3_column_bytes(sr.stmt.stmt, index);

            match len.cmp(&0) {
                std::cmp::Ordering::Equal => Ok(vec![]),
                std::cmp::Ordering::Less => Err(Error::InternalError(
                    "negative length returned from sqlite3_column_bytes",
                )),
                std::cmp::Ordering::Greater => {
                    Ok(std::slice::from_raw_parts(ptr.cast(), len as usize).to_vec())
                },
            }
        }
    }
}

#[cfg(test)]
mod sendsync_check {
    struct CheckSend<T: Send>(std::marker::PhantomData<T>);
    struct CheckSync<T: Sync>(std::marker::PhantomData<T>);

    #[test]
    fn check_send() {
        let _ = CheckSend::<super::Connection>(Default::default());
    }

    #[test]
    fn check_sync() {
        let _ = CheckSync::<super::Connection>(Default::default());
    }
}
