use crate::{
    db::{Connection, StatementContext},
    schema::{
        datum::Datum,
        entity::{helpers::is_relation, Entity, EntityID, EntityPart, EntityPartVisitor},
        relation::RelationData,
        Stored,
    },
    DBResult, Error,
};

use super::RelationNames;

pub(crate) fn insert<E: Entity>(conn: &Connection, value: &E) -> DBResult<E::ID> {
    struct InsertQuery<E: Entity>(std::marker::PhantomData<E>);

    conn.with_prepared(
        std::any::TypeId::of::<InsertQuery<E>>(),
        || {
            let mut part_names = String::new();
            let mut placeholders = String::new();
            struct PartNameVisitor<'a, E: Entity>(
                &'a mut String,
                &'a mut String,
                std::marker::PhantomData<E>,
            );
            impl<'a, E: Entity> EntityPartVisitor for PartNameVisitor<'a, E> {
                type Entity = E;
                fn visit<EP: EntityPart>(&mut self) {
                    // if this is a set-relation, then we don't actually want to do anything
                    // with it here; it doesn't have a column
                    if is_relation::<EP>() {
                        return;
                    }

                    if !self.0.is_empty() {
                        self.0.push_str(", ");
                        self.1.push_str(", ");
                    }
                    self.0.push('`');
                    self.0.push_str(EP::part_name());
                    self.0.push('`');
                    self.1.push('?');
                }
            }

            E::accept_part_visitor(&mut PartNameVisitor(
                &mut part_names,
                &mut placeholders,
                Default::default(),
            ));

            format!(
                "INSERT INTO `{}` ({}) VALUES ({}) RETURNING `id`",
                E::entity_name(),
                part_names,
                placeholders
            )
        },
        |mut ctx| {
            struct PartBinder<'a, 'b, E: Entity>(
                &'a mut StatementContext<'b>,
                i32,
                std::marker::PhantomData<E>,
            );
            impl<'a, 'b, E: Entity> EntityPartVisitor for PartBinder<'a, 'b, E> {
                type Entity = E;
                fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
                    // skip relations, as with the query preparation above
                    if is_relation::<EP>() {
                        return;
                    }

                    datum.bind_to(self.0, self.1);
                    self.1 += 1;
                }
            }

            value.accept_part_visitor_ref(&mut PartBinder(&mut ctx, 1, Default::default()));

            ctx.run()?
                .ok_or(Error::InternalError("No result row from INSERT query"))
                .map(|r| <E::ID>::from_raw(r.read(0).expect("couldn't read resulting ID")))
        },
    )
}

pub(crate) fn insert_and_return<E: Entity>(conn: &Connection, mut value: E) -> DBResult<Stored<E>> {
    let id = insert(conn, &value)?;

    // update relation data in all fields
    struct DatumWalker<'l, E: Entity>(&'l Connection, i64, std::marker::PhantomData<E>);
    impl<'l, E: Entity> EntityPartVisitor for DatumWalker<'l, E> {
        type Entity = E;
        fn visit_datum_mut<EP: EntityPart>(&mut self, datum: &mut EP::Datum) {
            datum.update_adata(RelationData {
                conn: self.0.clone(),
                part_name: EP::part_name(),
                local_name: <EP::Entity as Entity>::entity_name(),
                local_id: self.1,
            });
        }
    }

    value.accept_part_visitor_mut(&mut DatumWalker(conn, id.into_raw(), Default::default()));

    Ok(Stored::new(conn.clone(), id, value))
}

pub(crate) fn update_entity<E: Entity>(conn: &Connection, value: &Stored<E>) -> DBResult<()> {
    struct UpdateQuery<E: Entity>(std::marker::PhantomData<E>);

    conn.with_prepared(
        std::any::TypeId::of::<UpdateQuery<E>>(),
        || {
            let mut set_columns = String::new();
            struct PartNameVisitor<'a, E: Entity>(&'a mut String, std::marker::PhantomData<E>);
            impl<'a, E: Entity> EntityPartVisitor for PartNameVisitor<'a, E> {
                type Entity = E;
                fn visit<EP: EntityPart>(&mut self) {
                    // if this is a set-relation, then we don't actually want to do anything
                    // with it here; it doesn't have a column
                    if is_relation::<EP>() {
                        return;
                    }

                    if !self.0.is_empty() {
                        self.0.push_str(", ");
                    }
                    self.0.push('`');
                    self.0.push_str(EP::part_name());
                    self.0.push_str("` = ?");
                }
            }

            E::accept_part_visitor(&mut PartNameVisitor(&mut set_columns, Default::default()));
            format!(
                "UPDATE `{entity_name}` SET {set_columns} WHERE `id` = ?",
                entity_name = E::entity_name()
            )
        },
        |mut ctx| {
            struct PartBinder<'a, 'b, E: Entity>(
                &'a mut StatementContext<'b>,
                &'a mut i32,
                std::marker::PhantomData<E>,
            );
            impl<'a, 'b, E: Entity> EntityPartVisitor for PartBinder<'a, 'b, E> {
                type Entity = E;
                fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
                    // skip relations, as with the query preparation above
                    if is_relation::<EP>() {
                        return;
                    }

                    datum.bind_to(self.0, *self.1);
                    *self.1 += 1;
                }
            }

            // first bind all the updating clauses
            let mut index = 1;
            value.accept_part_visitor_ref(&mut PartBinder(
                &mut ctx,
                &mut index,
                Default::default(),
            ));

            // then bind the id
            value.id().bind_to(&mut ctx, index);

            ctx.run()?;

            Ok(())
        },
    )
}

pub(crate) fn do_connect<Remote: Entity>(
    rdata: &RelationData,
    an: RelationNames,
    remote_id: Remote::ID,
) -> DBResult<()> {
    rdata.conn.with_prepared(
        super::hash_of(("connect", an.local_name, an.remote_name, an.part_name)),
        || {
            format!(
                "insert into `{relation_name}` (`{local_field}`, `{remote_field}`) values (?, ?) returning (`id`)",
                relation_name = an.relation_name(),
                local_field = an.local_field,
                remote_field = an.remote_field
            )
        },
        |ctx| {
            ctx.bind(1, rdata.local_id)?;
            ctx.bind(2, remote_id.into_raw())?;

            ctx.run()?
                .ok_or(Error::ConstraintViolation("Relation entry uniqueness".to_string()))
                .map(|_| ())
        },
    )
}
