//! Component types for query construction.

use crate::{
    db::{Connection, StatementContext},
    prelude::Queryable,
    query::{QueryPart, RelationInterface},
    schema::{
        datum::{Datum, DatumDiscriminator, DatumVisitor, QueryEquivalent, QueryEquivalentList},
        entity::{Entity, EntityPart, EntityPartList, EntityPartVisitor},
        relation::{LocalSide, Relation},
        Stored,
    },
};

use super::{OutputContainer, Query};

/// Allow manipulation of an entire table.
pub(crate) struct TableComponent<E: Entity> {
    conn: Connection,
    _ghost: std::marker::PhantomData<E>,
}

impl<E: Entity> Clone for TableComponent<E> {
    fn clone(&self) -> Self {
        Self {
            conn: self.conn.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<E: Entity> TableComponent<E> {
    pub fn new(conn: Connection) -> Self {
        Self {
            conn,
            _ghost: Default::default(),
        }
    }
}

impl<E: Entity> Queryable for TableComponent<E> {
    type EntityOutput = E;
    type OutputContainer = Vec<Stored<E>>;
    type StaticVersion = Self;

    fn build(&self) -> Query {
        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT")
            .attach(QueryPart::Columns, "*")
            .attach(QueryPart::From, format!("`{}`", E::entity_name()))
    }
    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {}
    fn conn(&self) -> &Connection {
        &self.conn
    }
}

/// Filter on a Datum
#[derive(Clone)]
pub(crate) struct WithComponent<WEP: EntityPart, Parent: Queryable, QE: QueryEquivalent<WEP::Datum>>
{
    datum: QE,
    parent: Parent,
    _ghost: std::marker::PhantomData<WEP>,
}

impl<WEP: EntityPart, Parent: Queryable, QE: QueryEquivalent<WEP::Datum>>
    WithComponent<WEP, Parent, QE>
{
    pub fn new(parent: Parent, _part: WEP, datum: QE) -> Self {
        Self {
            datum,
            parent,
            _ghost: Default::default(),
        }
    }
}

/// this workaround is needed because we very explicitly would like QE to not be a
/// 'static-restricted type, and it doesn't matter for the purposes of actually distinguishing
/// between queries.
#[derive(Clone)]
pub(crate) struct CanonicalWithComponent<WEP: EntityPart, Parent: Queryable> {
    _ghost: std::marker::PhantomData<(WEP, Parent)>,
}

impl<WEP: EntityPart, Parent: Queryable + 'static> Queryable
    for CanonicalWithComponent<WEP, Parent>
{
    type EntityOutput = WEP::Entity;
    type OutputContainer = Option<Stored<WEP::Entity>>;
    type StaticVersion = Self;

    fn build(&self) -> Query {
        unreachable!()
    }
    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {
        unreachable!()
    }
    fn conn(&self) -> &Connection {
        unreachable!()
    }
}

impl<
        Parent: Queryable,
        WEP: EntityPart<Entity = Parent::EntityOutput>,
        QE: QueryEquivalent<WEP::Datum>,
    > Queryable for WithComponent<WEP, Parent, QE>
{
    type EntityOutput = WEP::Entity;
    type OutputContainer = Parent::OutputContainer;
    type StaticVersion = CanonicalWithComponent<WEP, Parent::StaticVersion>;

    fn build(&self) -> Query {
        self.parent.build().attach(
            QueryPart::Where,
            format!(
                "`{}`.`{}` = ?",
                <Self::EntityOutput>::entity_name(),
                WEP::part_name()
            ),
        )
    }
    fn bind(&self, stmt: &mut StatementContext, index: &mut i32) {
        self.parent.bind(stmt, index);
        self.datum.bind_to(stmt, *index);
        *index += 1;
    }

    fn conn(&self) -> &Connection {
        self.parent.conn()
    }
}

/// Filter on an index
pub(crate) struct IndexComponent<
    E: Entity,
    Parent: Queryable,
    EPL: EntityPartList<Entity = E>,
    EL: QueryEquivalentList<EPL::DatumList>,
> {
    datum: EL,
    parent: Parent,
    _ghost: std::marker::PhantomData<(E, EPL)>,
}

impl<
        E: Entity,
        Parent: Queryable,
        EPL: EntityPartList<Entity = E>,
        EL: QueryEquivalentList<EPL::DatumList>,
    > IndexComponent<E, Parent, EPL, EL>
{
    pub fn new(parent: Parent, datum: EL) -> Self {
        Self {
            datum,
            parent,
            _ghost: Default::default(),
        }
    }
}

impl<
        E: Entity,
        Parent: Queryable,
        EPL: EntityPartList<Entity = E>,
        EL: QueryEquivalentList<EPL::DatumList>,
    > Clone for IndexComponent<E, Parent, EPL, EL>
{
    fn clone(&self) -> Self {
        Self {
            datum: self.datum.clone(),
            parent: self.parent.clone(),
            _ghost: Default::default(),
        }
    }
}

/// this workaround is needed because we very explicitly would like EL to not be a
/// 'static-restricted type, and it doesn't matter for the purposes of actually distinguishing
/// between queries.
pub(crate) struct CanonicalIndexComponent<
    E: Entity,
    Parent: Queryable,
    EPL: EntityPartList<Entity = E>,
> {
    _ghost: std::marker::PhantomData<(E, Parent, EPL)>,
}

impl<E: Entity, Parent: Queryable + 'static, EPL: EntityPartList<Entity = E>> Queryable
    for CanonicalIndexComponent<E, Parent, EPL>
{
    type EntityOutput = E;
    type OutputContainer = Option<Stored<E>>;
    type StaticVersion = Self;

    fn build(&self) -> Query {
        unreachable!()
    }
    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {
        unreachable!()
    }
    fn conn(&self) -> &Connection {
        unreachable!()
    }
}

impl<E: Entity, Parent: Queryable + 'static, EPL: EntityPartList<Entity = E>> Clone
    for CanonicalIndexComponent<E, Parent, EPL>
{
    fn clone(&self) -> Self {
        Self {
            _ghost: Default::default(),
        }
    }
}

impl<
        E: Entity,
        Parent: Queryable,
        EPL: EntityPartList<Entity = E>,
        EL: QueryEquivalentList<EPL::DatumList>,
    > Queryable for IndexComponent<E, Parent, EPL, EL>
{
    type EntityOutput = E;
    type OutputContainer = Option<Stored<E>>;
    type StaticVersion = CanonicalIndexComponent<E, Parent::StaticVersion, EPL>;

    fn build(&self) -> Query {
        let mut query = self.parent.build();

        struct PartVisitor<'a, 'b, E: Entity>(&'a mut Query<'b>, std::marker::PhantomData<E>);
        impl<'a, 'b, E: Entity> EntityPartVisitor for PartVisitor<'a, 'b, E> {
            type Entity = E;
            fn visit<EP: EntityPart>(&mut self) {
                self.0.attach_mut(
                    QueryPart::Where,
                    format!(
                        "`{}`.`{}` = ?",
                        <EP::Entity>::entity_name(),
                        EP::part_name()
                    ),
                );
            }
        }

        EPL::accept_part_visitor(&mut PartVisitor::<E>(&mut query, Default::default()));

        query
    }

    fn bind(&self, stmt: &mut StatementContext, index: &mut i32) {
        self.parent.bind(stmt, index);

        struct Visitor<'a, 'b>(&'a mut StatementContext<'b>, &'a mut i32);
        impl<'a, 'b> DatumVisitor for Visitor<'a, 'b> {
            fn visit<ED: Datum>(&mut self, datum: &ED) {
                datum.bind_to(self.0, *self.1);
                *self.1 += 1;
            }
        }

        self.datum.accept(&mut Visitor(stmt, index));
    }

    fn conn(&self) -> &Connection {
        self.parent.conn()
    }
}

#[derive(Clone)]
pub(crate) struct SingleComponent<Parent: Queryable> {
    parent: Parent,
}

impl<Parent: Queryable> SingleComponent<Parent> {
    pub fn new(parent: Parent) -> Self {
        Self { parent }
    }
}

impl<Parent: Queryable> Queryable for SingleComponent<Parent> {
    type EntityOutput = Parent::EntityOutput;
    type OutputContainer = Option<Stored<Self::EntityOutput>>;
    type StaticVersion = SingleComponent<Parent::StaticVersion>;

    fn build(&self) -> Query {
        self.parent.build().attach(QueryPart::Trailing, "LIMIT 1")
    }

    fn bind(&self, stmt: &mut StatementContext, index: &mut i32) {
        self.parent.bind(stmt, index)
    }

    fn conn(&self) -> &Connection {
        self.parent.conn()
    }
}

/// Join with another entity via an relation
pub(crate) struct JoinComponent<R: Entity, L: Entity, EP: EntityPart<Entity = L>, Parent: Queryable>
{
    parent: Parent,
    _ghost: std::marker::PhantomData<(L, R, EP)>,
}

impl<R: Entity, L: Entity, EP: EntityPart<Entity = L>, Parent: Queryable>
    JoinComponent<R, L, EP, Parent>
{
    pub fn new(parent: Parent, _part: EP) -> Self {
        Self {
            parent,
            _ghost: Default::default(),
        }
    }
}

impl<R: Entity, L: Entity, EP: EntityPart<Entity = L>, Parent: Queryable> Clone
    for JoinComponent<R, L, EP, Parent>
{
    fn clone(&self) -> Self {
        Self {
            parent: self.parent.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<
        R: Entity,
        L: Entity,
        EP: EntityPart<Entity = L, Datum = AI>,
        AI: RelationInterface + Datum,
        Parent: Queryable,
    > Queryable for JoinComponent<R, L, EP, Parent>
{
    type EntityOutput = R;
    type OutputContainer = Vec<Stored<R>>;
    type StaticVersion = JoinComponent<R, L, EP, Parent::StaticVersion>;

    fn build(&self) -> Query {
        let remote_name = R::entity_name();
        let local_name = L::entity_name();

        struct Discriminator(Option<(&'static str, &'static str)>, Option<&'static str>);
        impl DatumDiscriminator for Discriminator {
            fn visit_entity_id<E: Entity>(&mut self) {
                unreachable!()
            }
            fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(&mut self) {
                unreachable!()
            }
            fn visit_bare_field<T: Datum>(&mut self) {
                unreachable!()
            }

            fn visit_relation_map<E: Entity>(&mut self) {
                self.0 = Some(("domain", "range"));
            }
            fn visit_relation_domain<R: Relation>(&mut self) {
                self.0 = Some(("domain", "range"));
                self.1 = Some(R::NAME);
            }
            fn visit_relation_range<R: Relation>(&mut self) {
                self.0 = Some(("range", "domain"));
                self.1 = Some(R::NAME);
            }
        }

        let mut d = Discriminator(None, None);
        <EP::Datum>::accept_discriminator(&mut d);

        let (local_field, remote_field) = d.0.unwrap();

        // table name format is {domain}_{range}_relation_{name}
        let relation_name = match AI::SIDE {
            LocalSide::Range => {
                format!(
                    "{remote_name}_{local_name}_relation_{}",
                    d.1.unwrap_or(EP::part_name())
                )
            },
            LocalSide::Domain => {
                format!(
                    "{local_name}_{remote_name}_relation_{}",
                    d.1.unwrap_or(EP::part_name())
                )
            },
        };

        self.parent
            .build()
            .attach(
                QueryPart::Join,
                format!(
                    "`{relation_name}` ON `{local_name}`.`id` = `{relation_name}`.`{local_field}`"
                ),
            )
            .attach(
                QueryPart::Join,
                format!(
                    "`{remote_name}` ON `{relation_name}`.`{remote_field}` = `{remote_name}`.`id`"
                ),
            )
            .replace(QueryPart::Columns, format!("`{remote_name}`.*"))
    }
    fn bind(&self, stmt: &mut StatementContext, index: &mut i32) {
        self.parent.bind(stmt, index);
    }

    fn conn(&self) -> &Connection {
        self.parent.conn()
    }
}

/// Get an entity via a foreign key.
pub(crate) struct ForeignComponent<FE: Entity, EP: EntityPart, Parent: Queryable> {
    parent: Parent,
    _ghost: std::marker::PhantomData<(FE, EP)>,
}

impl<FE: Entity, EP: EntityPart, Parent: Queryable> ForeignComponent<FE, EP, Parent> {
    pub fn new(parent: Parent, _part: EP) -> Self {
        Self {
            parent,
            _ghost: Default::default(),
        }
    }
}

impl<FE: Entity, EP: EntityPart, Parent: Queryable> Clone for ForeignComponent<FE, EP, Parent> {
    fn clone(&self) -> Self {
        Self {
            parent: self.parent.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<FE: Entity, EP: EntityPart, Parent: Queryable> Queryable for ForeignComponent<FE, EP, Parent> {
    type EntityOutput = FE;
    type StaticVersion = ForeignComponent<FE, EP, Parent::StaticVersion>;
    type OutputContainer =
        <Parent::OutputContainer as OutputContainer<Parent::EntityOutput>>::ReplacedEntity<FE>;

    fn build(&self) -> Query {
        let subquery = self.parent.build().replace(
            QueryPart::Columns,
            format!("`{}`.`id`", EP::Entity::entity_name()),
        );

        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT")
            .attach(QueryPart::Columns, "*")
            .attach(QueryPart::From, format!("`{}`", FE::entity_name()))
            .attach(
                QueryPart::Where,
                format!("`{}`.`id` = ({})", FE::entity_name(), subquery.assemble()),
            )
    }
    fn bind(&self, stmt: &mut StatementContext, index: &mut i32) {
        self.parent.bind(stmt, index)
    }
    fn conn(&self) -> &Connection {
        self.parent.conn()
    }
}
