//! Schema specification types.
//!
//! The following terminology used in some types in this module:
//! - *domain*: one side of a relation, the "pointed-from" side for one-sided relations
//! - *range*: one side of a relation, the "pointed-to" side for one-sided relations
//! - *local*: the current side of a relation
//! - *remote*: the opposite side of a relation

use query::Queryable;

use crate::{
    db::{Connection, StatementContext, StatementRow, Transaction},
    query::{self, Insertable},
    schema::datum::{Datum, DatumDiscriminator},
    schema::entity::Entity,
};
use crate::{DBResult, Error};

use self::{
    datum::{ConcreteDatum, DatumDiscriminatorRef},
    entity::EntityPartList,
};

/// Types related to datums, or struct fields.
pub mod datum;
/// Types related to entities, or structs that can be serialized into/deserialized from the
/// database.
pub mod entity;

/// Types related to inter-entity relationships.
pub mod relation;

/// Types related to indexes.
pub mod index;

mod build;
mod collect;
pub(crate) mod meta;

// ----------------------------------------------------------------------
// API types
// ----------------------------------------------------------------------

/// Wrapper struct that holds both an EntityID and an Entity itself.
pub struct Stored<T: Entity> {
    db: Connection,
    id: T::ID,
    wrap: T,
}

impl<T: Entity> Stored<T> {
    pub(crate) fn new(db: Connection, id: T::ID, value: T) -> Self {
        Self {
            db,
            id,
            wrap: value,
        }
    }

    /// Retrieve the entity ID of the stored entity.
    pub fn id(&self) -> T::ID {
        self.id
    }

    /// Consume the `Stored<>` wrapper and return the wrapped data.
    pub fn wrapped(self) -> T {
        self.wrap
    }

    /// Synchronize the wrapped value with the corresponding database row.
    pub fn sync(&mut self) -> DBResult<()> {
        let txn = Transaction::new(&self.db)?;
        query::base_queries::update_entity(&self.db, self)?;
        txn.commit()
    }
}

impl<T: Entity + std::fmt::Debug> std::fmt::Debug for Stored<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "Stored {{ id: {:?}, value: {:?} }}",
            self.id, self.wrap
        ))
    }
}

impl<T: Entity> AsRef<T> for Stored<T> {
    fn as_ref(&self) -> &T {
        &self.wrap
    }
}

impl<T: Entity> AsMut<T> for Stored<T> {
    fn as_mut(&mut self) -> &mut T {
        &mut self.wrap
    }
}

impl<T: Entity> std::ops::Deref for Stored<T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.wrap
    }
}

impl<T: Entity> std::ops::DerefMut for Stored<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.wrap
    }
}

impl<T: Entity> PartialEq for Stored<T> {
    fn eq(&self, other: &Self) -> bool {
        self.id == other.id
    }
}

// ----------------------------------------------------------------------
// Entity field types
// ----------------------------------------------------------------------

/// Stores an arbitrary Rust data type as serialized JSON in a string field.
#[derive(Clone)]
pub struct Serialized<T: serde::Serialize + serde::de::DeserializeOwned + Clone> {
    wrapped: T,
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Default + Clone> Default
    for Serialized<T>
{
    fn default() -> Self {
        Self {
            wrapped: T::default(),
        }
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone> std::fmt::Debug
    for Serialized<T>
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        <T as std::fmt::Debug>::fmt(&self.wrapped, f)
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> From<T> for Serialized<T> {
    fn from(value: T) -> Self {
        Self { wrapped: value }
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> AsRef<T> for Serialized<T> {
    fn as_ref(&self) -> &T {
        &self.wrapped
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> AsMut<T> for Serialized<T> {
    fn as_mut(&mut self) -> &mut T {
        &mut self.wrapped
    }
}

impl<T: 'static + serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone> Datum
    for Serialized<T>
{
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        let json = std::pin::Pin::new(
            serde_json::to_string(&self.wrapped).expect("couldn't serialize object into JSON"),
        );

        <&str as Datum>::bind_to(&&*json.as_ref(), stmt, index);

        // transfer ownership so that the data is still around for later
        stmt.transfer(json);
    }

    fn build_from(
        rdata: relation::RelationData,
        stmt: &mut StatementRow,
        index: &mut i32,
    ) -> DBResult<Self>
    where
        Self: Sized,
    {
        let s = <String as Datum>::build_from(rdata, stmt, index)?;

        let d = serde_json::from_str::<T>(s.as_str()).map_err(Error::JSON)?;

        Ok(Self { wrapped: d })
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_serialized::<T>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_serialized::<T>(&self.wrapped);
    }
}

impl<T: 'static + serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone>
    ConcreteDatum for Serialized<T>
{
}

// ----------------------------------------------------------------------
// Database specification types
// ----------------------------------------------------------------------

/// Table with EntityID-based lookup.
#[derive(Clone)]
pub struct IDMap<T: Entity> {
    pub(crate) conn: Connection,
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> IDMap<T> {
    /// Construct a non-empty instance of an `IDMap`.
    pub fn build(db: Connection) -> Self {
        Self {
            conn: db,
            _ghost: std::marker::PhantomData,
        }
    }

    pub(crate) fn conn(&self) -> &Connection {
        &self.conn
    }

    /// Look up an Entity in this map by ID.
    pub fn by_id(&self, id: T::ID) -> DBResult<Option<Stored<T>>> {
        self.with(id, &id).first().get()
    }
}

impl<T: Entity> Insertable<T> for IDMap<T> {
    /// Insert a new Entity into this map, and return its new ID.
    fn insert(&self, value: T) -> DBResult<T::ID> {
        let txn = Transaction::new(self.conn())?;
        let out = query::base_queries::insert(self.conn(), &value)?;
        txn.commit()?;
        Ok(out)
    }

    fn insert_and_return(&self, value: T) -> DBResult<Stored<T>> {
        let txn = Transaction::new(self.conn())?;
        let out = query::base_queries::insert_and_return(self.conn(), value)?;
        txn.commit()?;
        Ok(out)
    }
}

impl<E: Entity> DatabaseItem for IDMap<E> {
    fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor) {
        visitor.visit_idmap::<E>();
    }
}

/// Represents a single root-level table or index in a database.
pub trait DatabaseItem {
    /// Accept an entity visitor for entity discovery.
    fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor);
}

/// Visitor trait for iterating across the types in a [`Database`].
pub trait DatabaseItemVisitor {
    /// Visit an `IDMap<T>` type.
    fn visit_idmap<T: Entity>(&mut self)
    where
        Self: Sized;
    /// Visit an `Index<T, PL>` type.
    fn visit_index<T: Entity, PL: EntityPartList<Entity = T>>(&mut self)
    where
        Self: Sized;
}

/// A root structure for the database specification graph.
pub trait Database {
    /// Open the SQLite database at the given URI and return it as an instance of a schema.
    ///
    /// This function will attempt to create the database file if it is not present.
    fn open_path<U: AsRef<str>>(uri: U) -> DBResult<Self>
    where
        Self: Sized,
    {
        let conn = Connection::new(uri.as_ref())?;
        let schema = build::collect_from_database::<Self>();
        match schema.check(conn.clone()) {
            // schema checks out
            Some(true) => {},
            // schema doesn't match
            Some(false) => Err(Error::IncompatibleSchema)?,
            // no schema found
            None => {
                schema.create(conn.clone())?;
            },
        }

        conn.execute_raw_sql("PRAGMA foreign_keys = ON")?;

        Ok(Self::build(conn))
    }

    #[doc(hidden)]
    fn build(conn: Connection) -> Self
    where
        Self: Sized;

    /// Accept a visitor for iteration.
    fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor)
    where
        Self: Sized;
}
