use crate::{
    prelude::*,
    schema::{
        collect::{EntityStateContainer, PartType},
        entity::{Entity, EntityPart, EntityPartList, EntityPartVisitor, EntityVisitor},
        meta, Connection, DatabaseItemVisitor,
    },
    DBResult,
};

#[derive(Debug)]
struct ColumnInfo {
    name: &'static str,
    ty: String,
    fkey: Option<String>,
    unique: bool,
}

#[derive(Debug)]
struct TableInfo {
    table_name: String,
    columns: Vec<ColumnInfo>,
    constraints: Vec<String>,
}

impl TableInfo {
    fn new(name: String) -> Self {
        TableInfo {
            table_name: name,
            columns: vec![],
            constraints: vec![],
        }
    }

    fn build_creation_query(&self) -> String {
        let columns = self.columns.iter().map(|col| {
            format!(
                ", `{}` {}{}",
                col.name,
                col.ty,
                if col.unique { " unique" } else { "" }
            )
        });
        let fkeys = self.columns.iter().filter_map(|col| {
            Some(format!(
                ", foreign key(`{}`) references {} on delete cascade",
                col.name,
                col.fkey.as_ref()?
            ))
        });

        format!(
            "create table `{}` (`id` integer primary key{}{}{});",
            self.table_name,
            columns.collect::<String>(),
            fkeys.collect::<String>(),
            self.constraints
                .iter()
                .fold(String::new(), |a, b| format!("{}, {}", a, b))
        )
    }
}

#[derive(Debug, Hash)]
struct IndexInfo {
    table_name: String,
    columns: Vec<String>,
    unique: bool,
}

impl IndexInfo {
    fn build_creation_query(&self) -> String {
        let index_name = format!(
            "index_{}_{}",
            self.table_name,
            self.columns
                .iter()
                .cloned()
                .reduce(|a, b| format!("{a}_{b}"))
                .unwrap()
        );

        format!(
            "create {unique} index `{index_name}` on `{table_name}`({cols})",
            table_name = self.table_name,
            unique = if self.unique { "unique" } else { "" },
            cols = self.columns.join(",")
        )
    }
}

pub(crate) struct DatabaseSchema {
    signature: u64,
    queries: Vec<String>,
}

impl DatabaseSchema {
    const SCHEMA_SIGNATURE_KEY: &'static str = "schema_signature";

    /// Three possible results:
    /// - yes, this is a schema match (true)
    /// - no, this is not a schema match (false)
    /// - there is no schema that we know of (None)
    pub fn check(&self, db: Connection) -> Option<bool> {
        // attempt to use connection as a MetadataDB database
        let metadb = meta::MetadataDB::build(db);

        // check to see if the signature exists and matches
        metadb
            .metastore
            .keyed(&Self::SCHEMA_SIGNATURE_KEY.to_string())
            .get()
            .ok()
            .flatten()
            .map(|kv| kv.value.parse::<u64>().unwrap_or(0) == self.signature)
    }

    pub fn create(&self, db: Connection) -> DBResult<()> {
        db.execute_raw_sql("BEGIN TRANSACTION")?;
        for query in self.queries.iter() {
            log::trace!("Running creation query {query}");
            db.execute_raw_sql(query)?;
        }

        // attempt to use connection as a MetadataDB database
        let metadb = meta::MetadataDB::build(db.clone());

        for query in collect_from_database::<meta::MetadataDB>().queries.iter() {
            db.execute_raw_sql(query)?;
        }

        db.execute_raw_sql("COMMIT")?;

        // store signature
        metadb.metastore.insert(meta::Meta {
            key: Self::SCHEMA_SIGNATURE_KEY.into(),
            value: format!("{}", self.signature),
        })?;

        Ok(())
    }
}

pub(crate) fn collect_from_database<DB: Database>() -> DatabaseSchema {
    struct IV(EntityStateContainer, Vec<IndexInfo>);

    impl DatabaseItemVisitor for IV {
        fn visit_idmap<E: Entity>(&mut self)
        where
            Self: Sized,
        {
            self.0.visit::<E>();
        }

        fn visit_index<T: Entity, PL: EntityPartList<Entity = T>>(&mut self)
        where
            Self: Sized,
        {
            struct PV<E: Entity>(Vec<String>, std::marker::PhantomData<E>);
            impl<E: Entity> EntityPartVisitor for PV<E> {
                type Entity = E;

                fn visit<EP: EntityPart<Entity = Self::Entity>>(&mut self) {
                    self.0.push(EP::part_name().to_string());
                }
            }
            let mut pv = PV::<T>(Default::default(), Default::default());
            PL::accept_part_visitor(&mut pv);

            self.1.push(IndexInfo {
                table_name: T::entity_name().to_string(),
                columns: pv.0,
                unique: true,
            });
        }
    }

    let mut iv = IV(EntityStateContainer::default(), Default::default());

    DB::accept_item_visitor(&mut iv);

    // now to turn all that into a set of tables
    let mut tables = std::collections::HashMap::new();

    for state in iv.0.iter_states() {
        let table_name = state.name.to_string();
        // we may end up visiting duplicate entities; skip them if so
        if tables.contains_key(&table_name) {
            continue;
        }

        let mut table = TableInfo::new(table_name.clone());
        for part in state.parts.iter() {
            match &part.ty {
                PartType::Datum(dtype) => table.columns.push(ColumnInfo {
                    name: part.name,
                    ty: dtype.to_string(),
                    fkey: None,
                    unique: part.unique,
                }),
                PartType::IDReference(entity_name) => table.columns.push(ColumnInfo {
                    name: part.name,
                    ty: "int".into(),
                    fkey: Some(format!("`{}`(`id`)", entity_name)),
                    unique: part.unique,
                }),
                PartType::RelationDomain {
                    table_name: relation_table_name,
                    range_name,
                    injective,
                } => {
                    let mut relation_table = TableInfo::new(relation_table_name.clone());

                    relation_table.columns.push(ColumnInfo {
                        name: "domain",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", table_name)),
                        unique: false,
                    });

                    relation_table.columns.push(ColumnInfo {
                        name: "range",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", range_name)),
                        unique: *injective,
                    });

                    relation_table
                        .constraints
                        .push("unique(`range`, `domain`)".to_string());

                    tables.insert(relation_table_name.clone(), relation_table);
                },
                PartType::RelationRange {
                    table_name: relation_table_name,
                    domain_name,
                    injective,
                } => {
                    let mut relation_table = TableInfo::new(relation_table_name.clone());

                    relation_table.columns.push(ColumnInfo {
                        name: "domain",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", domain_name)),
                        unique: false,
                    });

                    relation_table.columns.push(ColumnInfo {
                        name: "range",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", table_name)),
                        unique: *injective,
                    });

                    relation_table
                        .constraints
                        .push("unique(`range`, `domain`)".to_string());
                    tables.insert(relation_table_name.clone(), relation_table);
                },
            }
        }

        let key = state.parts.iter().filter(|p| p.key).collect::<Vec<_>>();
        if !key.is_empty() {
            table.constraints.push(format!(
                "/* keying index */ unique({})",
                key.into_iter()
                    .map(|s| format!("`{}`", s.name))
                    .reduce(|a, b| format!("{},{}", a, b))
                    .unwrap()
            ));
        }

        tables.insert(table_name, table);
    }

    // this must be a stable hash function, so we very explicitly want to use a SipHasher with
    // known parameters
    #[allow(deprecated)]
    let mut signature_hasher = std::hash::SipHasher::new();
    use std::hash::{Hash, Hasher};

    let mut queries = vec![];

    for (table_name, table) in tables {
        let create_sql = table.build_creation_query();

        table_name.hash(&mut signature_hasher);
        create_sql.hash(&mut signature_hasher);
        queries.push(create_sql);
    }

    for iinfo in iv.1.into_iter() {
        iinfo.hash(&mut signature_hasher);
        queries.push(iinfo.build_creation_query());
    }

    DatabaseSchema {
        signature: signature_hasher.finish(),
        queries,
    }
}
