use crate::{
    db::{StatementContext, StatementRow},
    schema::entity::{Entity, EntityVisitor},
    schema::relation::{Relation, RelationData, RelationDomain, RelationMap, RelationRange},
    DBResult,
};

// trait implementations for Datum
mod datum_common;
// trait implementations for DatumList
mod datum_list;

// ----------------------------------------------------------------------
// Datum and related types
// ----------------------------------------------------------------------

/// Represents a data field in an Entity.
pub trait Datum: Clone + std::fmt::Debug {
    /// The type used in a SQL database to represent this datum.
    fn sql_type() -> &'static str;

    /// Generate a `std::fmt::Debug`-compatible view of the current datum; used to produce more
    /// human-friendly outputs for some more complicated types with state stored in the database.
    fn debug_field(&self, field: &'static str, fmt: &mut std::fmt::DebugStruct)
    where
        Self: Sized,
    {
        fmt.field(field, self);
    }

    /// Bind this datum to a [`StatementContext`] at a given index.
    fn bind_to(&self, _stmt: &mut StatementContext, index: i32);
    /// Construct an instance of this datum from a table row.
    fn build_from(
        _adata: RelationData,
        _stmt: &mut StatementRow,
        _index: &mut i32,
    ) -> DBResult<Self>
    where
        Self: Sized,
    {
        unreachable!()
    }

    /// Update any [`RelationData`] instances.
    #[doc(hidden)]
    fn update_adata(&mut self, _adata: RelationData) {}

    /// Accept an entity visitor to iterate across any entities this Datum type references.
    fn accept_entity_visitor(_: &mut impl EntityVisitor) {}
    /// Accept a datum discriminator without any instance reference to allow for more fine-grained per-type processing.
    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_bare_field::<Self>();
    }

    /// Accept a datum discriminator with an instance reference to allow for more fine-grained per-type processing.
    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_bare_field::<Self>(self);
    }
}

/// Marker trait for 'concrete' datums, i.e. those with a static lifetime. This separation is used
/// to ensure [`QueryEquivalent`] can be implemented in the current Rust type system.
pub trait ConcreteDatum: 'static + Datum {}

/// Visitor for allowing for type-specific behaviour.
pub trait DatumDiscriminator {
    /// Visit an `EntityID` type.
    fn visit_entity_id<E: Entity>(&mut self);
    /// Visit a `Serialized<T>` instance.
    fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(&mut self);
    /// Visit a bare datum field instance, such as a `String` or `usize`.
    fn visit_bare_field<T: Datum>(&mut self);
    /// Visit a one-direction relation map.
    fn visit_relation_map<E: Entity>(&mut self);
    /// Visit the domain side of a bidirectional relation map.
    fn visit_relation_domain<R: Relation>(&mut self);
    /// Visit the range side of a bidirectional relation map.
    fn visit_relation_range<R: Relation>(&mut self);
}

/// Visitor for allowing for type-specific behaviour, with instances passed as references.
pub trait DatumDiscriminatorRef {
    /// Visit an `EntityID` instance.
    fn visit_entity_id<E: Entity>(&mut self, _: &E::ID);
    /// Visit a `Serialized<T>` instance.
    fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(&mut self, _: &T);
    /// Visit a bare datum field instance, such as a `String` or `usize`.
    fn visit_bare_field<T: Datum>(&mut self, _: &T);
    /// Visit a one-direction relation map instance.
    fn visit_relation_map<E: Entity>(&mut self, _: &RelationMap<E>);
    /// Visit an instance of the domain side of a bidirectional relation map.
    fn visit_relation_domain<R: Relation>(&mut self, _: &RelationDomain<R>);
    /// Visit an instance of the range side of a bidirectional relation map.
    fn visit_relation_range<R: Relation>(&mut self, _: &RelationRange<R>);
}

/// A fixed-length list of EntityDatums, usually a tuple.
pub trait DatumList: Clone {
    /// Accept a datum visitor for iteration.
    fn accept(&self, visitor: &mut impl DatumVisitor);
}

/// A list of concrete datums.
pub trait ConcreteDatumList: DatumList + Clone {
    /// Construct a [`DatumList`] that is _query-equivalent_ of the current type. See
    /// [`QueryEquivalent`] and [`QueryEquivalentList`] for more information.
    fn build_equivalent<'l>(
        from: impl Iterator<Item = &'l str>,
    ) -> Option<impl QueryEquivalentList<Self> + 'l>;
}

/// A walker for DatumList instances.
pub trait DatumVisitor {
    /// Visit a specific datum in a DatumList.
    fn visit<ED: Datum>(&mut self, datum: &ED);
}

/// A forward-edge mapping between datums where one can stand in for the other during queries.
///
/// This is purely a marker trait, since the equivalence is based on how sqlite handles things.
/// Because this is a forward-edge mapping, if type `X` implements `QueryEquivalent<Y>`, it means `X` can
/// stand in for `Y` during a query. For example, the following should be true:
/// 1. for datum `T`, `&T` should implement `QueryEquivalent<T>`
///     - rationale: you should be able to query for something matching a value without passing an
///     owned value
/// 2. for datum `T`, `StringQuery` should implement `QueryEquivalent<T>`
///     - rationale: `StringQuery` explicitly shoves off equality checking to sqlite, which means
///     you should be able to use it in any query place.
///
/// Note that this is abstracted across [`ConcreteDatum`] and not [`Datum`]; this is to avoid
/// issues with [`StringQuery`] -- which is not concrete -- having two implementations of
/// `QueryEquivalent<StringQuery>`: one from the reflexive implementation and one from (2) above.
pub trait QueryEquivalent<T: ConcreteDatum>: Datum {}

// Every type is query-equivalent to itself.
impl<T: ConcreteDatum> QueryEquivalent<T> for T {}
// Every reference type is query-equivalent to the non-referenced type.
impl<'l, T: ConcreteDatum> QueryEquivalent<T> for &'l T {}

/// Used if you wish to use sqlite's built-in string equivalence for a query.
#[derive(Clone, Debug)]
pub struct StringQuery<'l>(pub &'l str);

/// A list version of [`QueryEquivalent`].
pub trait QueryEquivalentList<T: DatumList>: DatumList {}
