use crate::{
    db::{self, Bindable, StatementContext, StatementRow},
    schema::{datum::Datum, relation::RelationData},
    DBResult, Error,
};

use super::{
    ConcreteDatum, DatumDiscriminator, DatumDiscriminatorRef, QueryEquivalent, StringQuery,
};

impl<'l> Datum for StringQuery<'l> {
    fn sql_type() -> &'static str {
        "text"
    }
    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        self.0.bind_to(stmt, index)
    }
}

// a StringQuery and any datum are equivalent for the purposes of a query
// impl<'l, 'b, T: ConcreteDatum> QueryEquivalent<StringQuery<'l>> for &'b T {}
impl<'l, T: ConcreteDatum> QueryEquivalent<T> for StringQuery<'l> {}

impl ConcreteDatum for time::OffsetDateTime {}
impl Datum for time::OffsetDateTime {
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        let ts = self.unix_timestamp();
        ts.bind_to(stmt, index)
    }

    fn build_from(rdata: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let unix = i64::build_from(rdata, stmt, index)?;
        Self::from_unix_timestamp(unix).map_err(|e| Error::UnknownValue(e.to_string()))
    }
}

impl ConcreteDatum for String {}
impl Datum for String {
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, self.as_str())
            .expect("couldn't bind string");
    }

    fn build_from(_: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let val = stmt.read(*index)?;
        *index += 1;
        Ok(val)
    }
}

impl<'l> Datum for &'l str {
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, *self)
            .expect("couldn't bind string reference");
    }
}

// a str reference and an owned string are equivalent for the purposes of a query
impl<'l> QueryEquivalent<String> for &'l str {}

impl ConcreteDatum for usize {}
impl Datum for usize {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, *self as i64)
            .expect("couldn't bind usize as i64");
    }

    fn build_from(rdata: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(i64::build_from(rdata, stmt, index)? as usize)
    }
}

impl ConcreteDatum for isize {}
impl Datum for isize {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, *self as i64)
            .expect("couldn't bind isize as i64");
    }

    fn build_from(rdata: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(i64::build_from(rdata, stmt, index)? as isize)
    }
}

impl ConcreteDatum for u64 {}
impl Datum for u64 {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        todo!()
    }

    fn build_from(rdata: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(i64::build_from(rdata, stmt, index)? as u64)
    }
}

impl ConcreteDatum for i64 {}
impl Datum for i64 {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, *self).expect("couldn't bind i64")
    }

    fn build_from(_: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let val = stmt.read(*index)?;
        *index += 1;
        Ok(val)
    }
}

impl<T: 'static + Datum> ConcreteDatum for Option<T> {}
impl<T: Datum> Datum for Option<T> {
    fn sql_type() -> &'static str {
        T::sql_type()
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        if let Some(v) = self.as_ref() {
            v.bind_to(stmt, index);
        } else {
            // bind a NULL
            ().bind(stmt, index).expect("couldn't bind NULL for None");
        }
    }

    fn build_from(rdata: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        if stmt.read::<db::IsNull>(*index)?.0 {
            *index += 1;
            Ok(None)
        } else {
            T::build_from(rdata, stmt, index).map(Some)
        }
    }
}

impl ConcreteDatum for bool {}
impl Datum for bool {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, if *self { 1i64 } else { 0i64 })
            .expect("couldn't bind bool");
    }

    fn build_from(rdata: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(i64::build_from(rdata, stmt, index)? != 0)
    }
}

impl ConcreteDatum for Vec<u8> {}
impl Datum for Vec<u8> {
    fn sql_type() -> &'static str {
        "blob"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, self.as_slice())
            .expect("couldn't bind Vec<u8>");
    }

    fn build_from(_: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let val = stmt.read(*index)?;
        *index += 1;
        Ok(val)
    }
}

impl<'l> Datum for &'l [u8] {
    fn sql_type() -> &'static str {
        "blob"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, *self).expect("couldn't bind byte slice");
    }
}

// a byte slice and an owned byte slice are equivalent for the purposes of a query
impl<'l> QueryEquivalent<Vec<u8>> for &'l [u8] {}

impl Datum for f32 {
    fn sql_type() -> &'static str {
        "numeric"
    }

    fn build_from(_adata: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let val = stmt.read(*index)?;
        *index += 1;
        Ok(val)
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, *self).expect("couldn't bind f32");
    }
}
impl ConcreteDatum for f32 {}

impl Datum for f64 {
    fn sql_type() -> &'static str {
        "numeric"
    }

    fn build_from(_adata: RelationData, stmt: &mut StatementRow, index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        let val = stmt.read(*index)?;
        *index += 1;
        Ok(val)
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        stmt.bind(index, *self).expect("couldn't bind f64");
    }
}
impl ConcreteDatum for f64 {}

// Generic Datum implementation for references to a Datum
impl<'l, T: Datum> Datum for &'l T {
    fn sql_type() -> &'static str {
        T::sql_type()
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        T::bind_to(self, stmt, index)
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        T::accept_discriminator(d)
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        T::accept_discriminator_ref(self, d)
    }

    fn accept_entity_visitor(v: &mut impl crate::schema::entity::EntityVisitor) {
        T::accept_entity_visitor(v)
    }
}
