use crate::{
    db::{Connection, StatementContext, StatementRow},
    query::RelationInterface,
    schema::{
        datum::{ConcreteDatum, Datum, DatumDiscriminator, DatumDiscriminatorRef},
        entity::{Entity, EntityVisitor},
    },
    DBResult, Error,
};

/// Represents an arbitrary relation between two types of entities.
///
/// In its most generic form, this represents a table that stores an arbitrary set of `(Domain::ID,
/// Range::ID)` pairs used to define the relation. Can be restricted to be injective if this is
/// desired. Doing so will incur a small runtime cost, since an extra index must be maintained.
pub trait Relation: 'static {
    /// The domain of the relation, aka the "pointed-from" type. This is interchangeable with
    /// `Range` unless `INJECTIVE` is set to true.
    type Domain: Entity;
    /// The range of the relation, aka the "pointed-to" type. This is interchangeable with
    /// `Domain` unless `INJECTIVE` is set to true.
    type Range: Entity;

    /// If true, then each Range-type entity can only be referred to by a single Domain-type
    /// entity. This roughly translates to the relation being computationally-efficiently
    /// invertible for a two-way map.
    const INJECTIVE: bool = false;

    /// A unique, human-readable name for the relation, which should be short and not include
    /// spaces.
    const NAME: &'static str;
}

/// Enumeration used to represent which side of a relation an [`RelationInterface`] trait implementation is representing.
#[derive(Debug)]
pub enum LocalSide {
    /// The side of the relation that the [`RelationInterface`] represents is the `Domain` side.
    Domain,
    /// The side of the relation that the [`RelationInterface`] represents is the `Range` side.
    Range,
}

/// Opaque data structure used for constructing `Relation{Map,Domain,Range}` instances.
#[derive(Clone)]
pub struct RelationData {
    pub(crate) conn: Connection,
    pub(crate) local_name: &'static str,
    pub(crate) part_name: &'static str,
    pub(crate) local_id: i64,
}

/// Represents a simple one-to-many non-injective entity relationship.
///
/// This is a 'shortcut' type that doesn't require defining a tag type that implements `Relation`.
pub struct RelationMap<T: Entity> {
    pub(crate) data: Option<RelationData>,
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> Clone for RelationMap<T> {
    fn clone(&self) -> Self {
        Self {
            data: self.data.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<T: Entity> std::fmt::Debug for RelationMap<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "RelationMap {{ id: {:?} }}",
            self.data.as_ref().map(|d| d.local_id)
        ))
    }
}

impl<T: Entity> Default for RelationMap<T> {
    fn default() -> Self {
        Self {
            data: None,
            _ghost: Default::default(),
        }
    }
}

impl<T: Entity> RelationInterface for RelationMap<T> {
    type RemoteEntity = T;
    const SIDE: LocalSide = LocalSide::Domain;

    fn get_distinguishing_name(&self) -> DBResult<&'static str> {
        self.data
            .as_ref()
            .ok_or(Error::LogicError(
                "no distinguishing name for empty RelationMap",
            ))
            .map(|d| d.part_name)
    }

    fn get_data(&self) -> DBResult<&RelationData> {
        self.data
            .as_ref()
            .ok_or(Error::LogicError("Reading from unassigned RelationMap"))
    }
}

impl<T: Entity> Datum for RelationMap<T> {
    fn sql_type() -> &'static str {
        unreachable!()
    }

    fn debug_field(&self, _field: &'static str, _fmt: &mut std::fmt::DebugStruct)
    where
        Self: Sized,
    {
    }

    fn accept_entity_visitor(v: &mut impl EntityVisitor) {
        v.visit::<T>();
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_relation_map::<T>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_relation_map::<T>(self);
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        unreachable!()
    }

    fn build_from(rdata: RelationData, _stmt: &mut StatementRow, _index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(Self {
            data: Some(rdata),
            _ghost: Default::default(),
        })
    }

    fn update_adata(&mut self, rdata: RelationData) {
        self.data = Some(rdata);
    }
}
impl<T: Entity> ConcreteDatum for RelationMap<T> {}

/// The domain part of a full Relation definition.
pub struct RelationDomain<R: Relation> {
    pub(crate) data: Option<RelationData>,
    _ghost: std::marker::PhantomData<R>,
}

impl<R: Relation> Clone for RelationDomain<R> {
    fn clone(&self) -> Self {
        Self {
            data: self.data.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<R: Relation> Default for RelationDomain<R> {
    fn default() -> Self {
        Self {
            data: None,
            _ghost: Default::default(),
        }
    }
}

impl<R: Relation> std::fmt::Debug for RelationDomain<R> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "RelationDomain {{ id: {:?} }}",
            self.data.as_ref().map(|d| d.local_id)
        ))
    }
}

impl<R: Relation> RelationInterface for RelationDomain<R> {
    type RemoteEntity = R::Range;
    const SIDE: LocalSide = LocalSide::Domain;

    fn get_distinguishing_name(&self) -> DBResult<&'static str> {
        Ok(R::NAME)
    }

    fn get_data(&self) -> DBResult<&RelationData> {
        self.data
            .as_ref()
            .ok_or(Error::LogicError("Reading from unassigned RelationDomain"))
    }
}

impl<R: Relation> Datum for RelationDomain<R> {
    fn sql_type() -> &'static str {
        unreachable!()
    }

    fn debug_field(&self, _field: &'static str, _fmt: &mut std::fmt::DebugStruct)
    where
        Self: Sized,
    {
    }

    fn accept_entity_visitor(v: &mut impl EntityVisitor) {
        v.visit::<R::Range>();
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_relation_domain::<R>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_relation_domain(self);
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        unreachable!()
    }

    fn build_from(rdata: RelationData, _stmt: &mut StatementRow, _index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(Self {
            data: Some(rdata),
            _ghost: Default::default(),
        })
    }

    fn update_adata(&mut self, rdata: RelationData) {
        self.data = Some(rdata);
    }
}
impl<R: Relation> ConcreteDatum for RelationDomain<R> {}

/// The range part of a full Relation definition.
pub struct RelationRange<R: Relation> {
    pub(crate) data: Option<RelationData>,
    _ghost: std::marker::PhantomData<R>,
}

impl<R: Relation> Clone for RelationRange<R> {
    fn clone(&self) -> Self {
        Self {
            data: self.data.clone(),
            _ghost: Default::default(),
        }
    }
}

impl<R: Relation> Default for RelationRange<R> {
    fn default() -> Self {
        Self {
            data: None,
            _ghost: Default::default(),
        }
    }
}

impl<R: Relation> std::fmt::Debug for RelationRange<R> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "RelationRange {{ id: {:?} }}",
            self.data.as_ref().map(|d| d.local_id)
        ))
    }
}

impl<R: Relation> RelationInterface for RelationRange<R> {
    type RemoteEntity = R::Domain;
    const SIDE: LocalSide = LocalSide::Range;

    fn get_distinguishing_name(&self) -> DBResult<&'static str> {
        Ok(R::NAME)
    }

    fn get_data(&self) -> DBResult<&RelationData> {
        self.data
            .as_ref()
            .ok_or(Error::LogicError("Reading from unassigned RelationRange"))
    }
}

impl<R: Relation> Datum for RelationRange<R> {
    fn sql_type() -> &'static str {
        unreachable!()
    }

    fn debug_field(&self, _field: &'static str, _fmt: &mut std::fmt::DebugStruct)
    where
        Self: Sized,
    {
    }

    fn accept_entity_visitor(v: &mut impl EntityVisitor) {
        v.visit::<R::Domain>();
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_relation_range::<R>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_relation_range(self);
    }

    fn bind_to(&self, _stmt: &mut StatementContext, _index: i32) {
        unreachable!()
    }

    fn build_from(rdata: RelationData, _stmt: &mut StatementRow, _index: &mut i32) -> DBResult<Self>
    where
        Self: Sized,
    {
        Ok(Self {
            data: Some(rdata),
            _ghost: Default::default(),
        })
    }

    fn update_adata(&mut self, rdata: RelationData) {
        self.data = Some(rdata);
    }
}
impl<R: Relation> ConcreteDatum for RelationRange<R> {}
